import json
from abc import ABC
from typing import Optional

from genie_flow_invoker.genie import GenieInvoker
from genie_flow_invoker.invoker.ms_sql_server.query import Querier
from genie_flow_invoker.invoker.ms_sql_server.retrieve import Retriever
from genie_flow_invoker.invoker.ms_sql_server.store import StoreStrategyType, Storer
from genie_flow_invoker.utils import get_config_value


class AbstractMSSQLServerInvoker(GenieInvoker, ABC):

    def __init__(
            self,
            server: str,
            username: str,
            password: str,
            database: str,
    ):
        """
        An abstract base class to store basic properties used in all subsequent MS SQL invokers.

        :param server: the host name of the server
        :param username: the username to use for login
        :param password: the password to use for login
        :param database: the database to use
        """

        self.base_config = dict(
            server=server,
            username=username,
            password=password,
            database=database,
        )

    @classmethod
    def _get_operator_config(cls, config: dict) -> dict:
        """
        Creates a new instance with the relevant parameters. All parameters are read from
        the configuration. The following parameters are also read from environment variables,
        where the environment variable takes precedence.

        - server MS_SQL_SERVER
        - username MS_SQL_SERVER_USERNAME
        - password MS_SQL_SERVER_PASSWORD
        - database MS_SQL_SERVER_DATABASE

        :param config: the configuration dictionary
        :return: a new instance with the relevant parameters
        """
        return dict(
            server=get_config_value(
                config,
                "MS_SQL_SERVER",
                "server",
                "SQL Server hostname",
            ),
            username=get_config_value(
                config,
                "MS_SQL_SERVER_USERNAME",
                "username",
                "SQL Server username",
            ),
            password=get_config_value(
                config,
                "MS_SQL_SERVER_PASSWORD",
                "password",
                "SQL Server password",
            ),
            database = get_config_value(
                config,
                "MS_SQL_SERVER_DATABASE",
                "database",
                "SQL Server database name",
            ),
        )

    def compile_config(self, content: str) -> dict:
        """
        Compiles the ultimate configuration into a dictionary. Precedence of configuration
        parameters is as follows:
        1. any parameter specified in the content passed to the invoker
        2. any parameter that is specified through an environment variable
        3. any parameter that is specified in the `meta.yaml` file

        :param content: the content that is passed to the invoker
        :return: a dictionary with the ultimate valued for the relevant parameters
        """
        try:
            override_config = json.loads(content)
        except json.decoder.JSONDecodeError:
            override_config = {}

        config = self.base_config.copy()
        config.update(override_config)
        return config


class AbstractMSSQLServerWithTableInvoker(AbstractMSSQLServerInvoker, ABC):

    def __init__(
            self,
            server: str,
            username: str,
            password: str,
            database: str,
            table: str,
    ):
        """
        An abstract base class to store basic properties used in all subsequent MS SQL invokers
        that refer to a table.

        :param server: the host name of the server
        :param username: the username to use for login
        :param password: the password to use for login
        :param database: the database to use
        :param table: the name of the table to retrieve data from
        """
        super().__init__(server, username, password, database)
        self.base_config["table"] = table

    @classmethod
    def _get_operator_config(cls, config: dict) -> dict:
        result = super()._get_operator_config(config)
        result["table"] = get_config_value(
            config,
            "MS_SQL_SERVER_TABLE",
            "table",
            "SQL Server table name",
        )
        return result


class MSSQLServerRetrieveInvoker(AbstractMSSQLServerWithTableInvoker):

    def __init__(
            self,
            server: str,
            username: str,
            password: str,
            database: str,
            table: str,
            top: Optional[int] = None,
            sort_order: Optional[list[str]] = None,
    ):
        """
        Invoker for retrieving data from an MSSQL server table.

        If `top` is specified, a maximum of `top` records will be returned.

        If `sort_order` is specified, the retrieved records will be sorted. That parameter
        should be a list of column names, in descending order of precedence. Putting a minus
        sign in front of the column name will do a descending order for that column.

        :param server: the host name of the server
        :param username: the username to use for login
        :param password: the password to use for login
        :param database: the database to use
        :param table: the name of the table to retrieve data from
        :param top: the maximum number of records to return
        :param sort_order: the column names to sort by
        """
        super().__init__(server, username, password, database, table)
        self.base_config["top"] = top
        self.base_config["sort_order"] = sort_order


    @classmethod
    def from_config(cls, config: dict):
        """
        Creates a new instance with the relevant parameters. All parameters are read from the
        configuration. The following parameters are also used from environment variables:
        - server MS_SQL_SERVER
        - username MS_SQL_SERVER_USERNAME
        - password MS_SQL_SERVER_PASSWORD
        - database MS_SQL_SERVER_DATABASE
        - table MS_SQL_SERVER_TABLE

        :param config: a dictionary with configuration parameters
        :return: a new instance with the relevant parameters
        """
        operator_config = cls._get_operator_config(config)
        operator_config["top"]=config.get("top", None)
        operator_config["sort_order"]=config.get("sort_order", None)
        return cls(**operator_config)

    def invoke(self, content: str) -> str:
        """
        Invoke the retrieval. All parameters can be overriden in the content by specifying
        a JSON string that gives a new value to a configuration parameter.

        :param content: potentially a JSON string with overriding values
        :return: a JSON version of the retrieved records
        """
        config = self.compile_config(content)
        retriever = Retriever(**config)
        result = retriever.retrieve()

        return result.model_dump_json()


class MSSQLServerStoreInvoker(AbstractMSSQLServerWithTableInvoker):

    def __init__(
            self,
            server: str,
            username: str,
            password: str,
            database: str,
            table: str,
            strategy: StoreStrategyType = "insert",
            primary_key: list[str] = None,
    ):
        """
        An Invoker that can store data into a table, using different strategies.

        Specifying the storage strategy as "insert" (the default), "update" or "upsert". The
        first just adds the records, the second one updates existing records, the third one
        updates existing records but adds records when they do not yet exist.

        For the "update" and "upsert" strategies to work, a primary key must be provided.

        The primary key is specified as a list of column names.

        :param server: the host name of the server
        :param username: the username to use for login
        :param password: the password to use for login
        :param database: the database to use
        :param table: the name of the table to store data into
        :param strategy: the strategy to use ("insert", "update", "upsert")
        :param primary_key: the list of columns that make up the primary key
        """
        super().__init__(server, username, password, database, table)
        self.base_config["strategy"] = strategy
        self.base_config["primary_key"] = primary_key

    @classmethod
    def from_config(cls, config: dict):
        """
        Creates a new instance with the relevant parameters. All parameters are read from the
        configuration. The following parameters will also be read from environment variables,
        where the value specified in the environment variable takes precedence.

        - server MS_SQL_SERVER
        - username MS_SQL_SERVER_USERNAME
        - password MS_SQL_SERVER_PASSWORD
        - database MS_SQL_SERVER_DATABASE
        - table MS_SQL_SERVER_TABLE

        :param config: a dictionary with configuration parameters
        :return: a new instance with the relevant parameters
        """
        operator_config = super()._get_operator_config(config)
        operator_config["strategy"] = config.get("strategy", None)
        operator_config["primary_key"] = config.get("primary_key", None)
        return cls(**operator_config)

    def invoke(self, content: str) -> str:
        """
        Invoke the storing of a new record or set of records.

        Input expected is a JSON that contains an object with keys that are the column names
        and values for these columns to store. Optionally, this can be a list of such objects.

        Output is a JSON version of a list of statuses, one per record and in the same order.
        Status can be "inserted" or "updated", a warning or error message.

        :param content: the JSON encoded record or set of records
        :return: a JSON encoded list of statuses, one per record and in the same order
        """
        try:
            records = json.loads(content)
        except json.decoder.JSONDecodeError:
            records = []

        if not isinstance(records, list):
            records = [records]
        storer = Storer(**self.base_config)
        result = storer.store(records)

        return json.dumps(result)


class MSSQLServerQueryInvoker(AbstractMSSQLServerInvoker):

    @classmethod
    def from_config(cls, config: dict):
        """
        Creates a new instance with the relevant parameters. All parameters are read from
        the configuration. The following parameters are also read from environment variables,
        where the environment variable takes precedence.

        - server MS_SQL_SERVER
        - username MS_SQL_SERVER_USERNAME
        - password MS_SQL_SERVER_PASSWORD
        - database MS_SQL_SERVER_DATABASE

        :param config: the configuration dictionary
        :return: a new instance with the relevant parameters
        """
        config = cls._get_operator_config(config)
        return cls(**config)

    def invoke(self, content: str) -> str:
        """
        Send the content as a query to the configured MS SQL server. Returns a JSON encoding
        of a retrieved records object.

        :param content: the query to run
        :return: a JSON encoded retrieved records object
        """
        querier = Querier(**self.base_config)
        result = querier.query(content)
        return result.model_dump_json()



from hashlib import md5

import pymssql
from loguru import logger

from .manager import MSSQLServerOperator
from .model import RetrievedResults


class Querier(MSSQLServerOperator):

    def query(self, query: str) -> RetrievedResults:
        """
        Run the provided query against the configured database.

        :param query: the query to run
        :return: A `RetrievedResults` instance that contains the results and/or
        any errors or warnings that were raised.
        """
        logger.debug(
            "Sending query to MS SQL Server: '{}'",
            query,
        )
        logger.info(
            "Sending query to MS SQL Server: hashed '{}'",
            md5(query.encode('utf-8'), usedforsecurity=False).hexdigest(),
        )
        with self._manager_factory.create_manager() as connection:
            with connection.cursor(as_dict=True) as cursor:
                try:
                    cursor.execute(query)
                except pymssql.Error as e:
                    logger.warning("Query failed with error {}", str(e))
                    return RetrievedResults(
                        error=f"error {e.__class__.__name__}: {e.args}",
                    )
                except pymssql.Warning as e:
                    logger.warning("Query failed with warning {}", str(e))
                    return RetrievedResults(
                        results=cursor.fetchall() if cursor.rowcount > 0 else [],
                        error=f"warning {e.__class__.__name__}: {e.args}",
                    )

                logger.info("Query succeeded and returned {} rows", cursor.rowcount)
                return RetrievedResults(
                    results=cursor.fetchall(),
                )



import os

import pytest
from pytest import fixture
import pymssql

from genie_flow_invoker.invoker.ms_sql_server.query import Querier
from genie_flow_invoker.invoker.ms_sql_server.retrieve import Retriever
from genie_flow_invoker.invoker.ms_sql_server.store import Storer


if os.environ.get("SKIP_SQL_SERVER_TESTS") == "true":
    # We seem to have an issue with running these tests on the Gitlab Runner
    pytest.skip("Skipping SQLServer tests", allow_module_level=True)


@fixture(scope="function")
def connection(sql_server_details):
    connection = pymssql.connect(*sql_server_details)

    with connection.cursor() as cursor:
        cursor.execute("DROP TABLE IF EXISTS tesy")
    connection.commit()

    with connection.cursor() as cursor:
        cursor.execute("CREATE TABLE tesy (aap int, noot varchar(64))")
    connection.commit()

    with connection.cursor() as cursor:
        cursor.execute("INSERT INTO tesy VALUES (1, 'wim'), (2, 'zus'), (3, 'jet')")
    connection.commit()

    yield connection

    connection.close()


def test_sqlserver_retrieve(sql_server_details, connection):
    retriever = Retriever(*sql_server_details, table="tesy")
    results = retriever.retrieve()

    assert results.results == [
        dict(aap=1, noot="wim"),
        dict(aap=2, noot="zus"),
        dict(aap=3, noot="jet"),
    ]
    assert results.error is None


def test_sqlserver_retrieve_missing_table(sql_server_details, connection):
    retriever = Retriever(*sql_server_details, table="non_existent")
    results = retriever.retrieve()

    assert results.error is not None


def test_sqlserver_retrieve_ordered(sql_server_details, connection):
    retriever = Retriever(*sql_server_details, table="tesy", sort_order=["-aap"])
    results = retriever.retrieve()

    assert results.error is None
    assert results.results == [
        dict(aap=3, noot="jet"),
        dict(aap=2, noot="zus"),
        dict(aap=1, noot="wim"),
    ]


def test_sqlserver_store(sql_server_details, connection):
    storer = Storer(*sql_server_details, table="tesy")
    results = storer.store([dict(aap=10, noot="vuur")])

    with connection.cursor(as_dict=True) as cursor:
        cursor.execute("SELECT * FROM tesy WHERE aap = 10")
        vuur = cursor.fetchall()

    assert results == ["inserted"]
    assert vuur == [dict(aap=10, noot="vuur")]


def test_sqlserver_update(sql_server_details, connection):
    storer = Storer(
        *sql_server_details,
        table="tesy",
        strategy="update",
        primary_key=["aap"],
    )
    results = storer.store([dict(aap=1, noot="vuur")])

    with connection.cursor(as_dict=True) as cursor:
        cursor.execute("SELECT * FROM tesy WHERE aap = 1")
        vuur = cursor.fetchall()

    assert results == ["updated"]
    assert vuur == [dict(aap=1, noot="vuur")]


def test_sqlserver_upsert(sql_server_details, connection):
    storer = Storer(
        *sql_server_details,
        table="tesy",
        strategy="upsert",
        primary_key=["aap"],
    )
    results = storer.store([dict(aap=1, noot="vuur"), dict(aap=10, noot="vuur")])

    with connection.cursor(as_dict=True) as cursor:
        cursor.execute("SELECT * FROM tesy WHERE noot = 'vuur'")
        vuur = cursor.fetchall()

    assert results == ["updated", "inserted"]
    assert vuur == [dict(aap=1, noot="vuur"), dict(aap=10, noot="vuur")]


def test_sqlserver_query(sql_server_details, connection):
    querier = Querier(*sql_server_details)
    results = querier.query("SELECT MAX(aap) AS max_aap FROM tesy;")

    assert results.error is None
    assert results.results == [dict(max_aap=3)]



from abc import ABC
from typing import Optional

from loguru import logger
import pymssql


class MSSQLServerManager:

    def __init__(
            self,
            server: str,
            username: str,
            password: str,
            database: str,
    ):
        """
        A connection creator that takes server, username, password and database name and serves
        as a context manager for connections to MS SQL Server.

        :param server: the host name of the server to connect to
        :param username: the username to use
        :param password: the password to use
        :param database: the database to connect to
        """
        self._server = server
        self._username = username
        self._password = password
        self._database = database
        self._connection: Optional[pymssql.Connection] = None

    def __enter__(self):
        if self._connection is not None:
            logger.critical("Entering a context manager that is already entered")
            raise RuntimeError("Entering a context manager that is already entered")

        logger.info(
            "Creating connection to server {} with username {}",
            self._server,
            self._username
        )
        try:
            self._connection = pymssql.connect(
                self._server,
                self._username,
                self._password,
                self._database,
            )
        except Exception as e:
            logger.exception(e)
            raise

        return self._connection

    def __exit__(self, exc_type, exc_val, exc_tb):
        logger.info(
            "Closing connection to server {} with username {}",
            self._server,
            self._username,
        )
        try:
            self._connection.close()
        except Exception as e:
            logger.exception(e)
            raise


class MSSQLServerManagerFactory:
    """
    A factory for `MSSQLServerManager`s
    """

    def __init__(
            self,
            server: str,
            username: str,
            password: str,
            database: str,
    ):
        """
        Retain the base parameters required to create an MS SQL Server manager.

        :param server: the host name of the server to connect to
        :param username: the username to use
        :param password: the password to use
        :param database: the database to connect to
        """
        self._server = server
        self._username = username
        self._password = password
        self._database = database

    def create_manager(self) -> MSSQLServerManager:
        """
        Creates a new MS SQL Server manager instance with the persisted parameters from
        this factory.

        :return: a new instance of an `MSSQLServerManager`
        """
        return MSSQLServerManager(
            self._server,
            self._username,
            self._password,
            self._database,
        )


class MSSQLServerOperator(ABC):

    def __init__(
            self,
            server: str,
            username: str,
            password: str,
            database: str,
    ):
        """
        An abstract base class for MS SQL Server operators. Stores the base
        properties required to connect to the database server.

        Creates a manager factory able to create `MSSQLServerManager`s.

        :param server: the host name of the server to connect to
        :param username: the username to connect with
        :param password: the password to connect with
        :param database: the database to connect to
        """
        self._manager_factory = MSSQLServerManagerFactory(
            server,
            username,
            password,
            database,
        )



import pymssql.exceptions
import pytest

from genie_flow_invoker.invoker.ms_sql_server.manager import MSSQLServerManagerFactory
from conftest import MockMSSQLConnection


def test_connection_fails(monkeypatch):

    def connect(*args, **kwargs):
        raise pymssql.exceptions.DatabaseError("Failed to connect")

    monkeypatch.setattr(pymssql, "connect", connect)

    factory = MSSQLServerManagerFactory(
        "localhost",
        "sa",
        "FakePassword",
        "some_database",
    )
    with pytest.raises(pymssql.exceptions.DatabaseError) as excinfo:
        with factory.create_manager():
            ...

    assert "Failed to connect" in str(excinfo.value)


def test_connection_closes(monkeypatch):

    def connect(*args, **kwargs):
        return MockMSSQLConnection()

    monkeypatch.setattr(pymssql, "connect", connect)

    factory = MSSQLServerManagerFactory(
        "localhost",
        "sa",
        "FakePassword",
        "some_database",
    )
    with factory.create_manager() as connection:
        assert connection.open

    assert not connection.open



import collections
from collections import namedtuple
from typing import Literal, Any

from loguru import logger
import pymssql

from .manager import MSSQLServerOperator

StoreStrategyType = Literal["insert", "update", "upsert"]
IndexedRecord = namedtuple("IndexedRecord", ["index", "record"])


def quote_if_string(value: Any) -> Any:
    """
    Strings in SQL command needs to be quoted. This function will quote the passed value
    if it is a string.

    :param value: the value to (potentially) quote
    :return: the value as it can be used in a SQL command
    """
    if isinstance(value, str):
        return f"'{value}'"
    return value


class Storer(MSSQLServerOperator):

    def __init__(
            self,
            server: str,
            username: str,
            password: str,
            database: str,
            table: str,
            strategy: StoreStrategyType = "insert",
            primary_key: list[str] = None,
    ):
        """
        A Storer is an MS SQL Server operator that stores records in a table. Depending on the
        strategy chose, it "insert"s, "update"s or "upsert"s new records. The output is a
        list of results. These can be "inserted", "updated" or a warning or error, followed by
        the message sent from the server.

        :param server: the hostname of the server
        :param username: the username to login to the server
        :param password: the password to login to the server
        :param database: the database to use when logged in
        :param table: the table to store into
        :param strategy: the strategy to use ("insert", "update", "upsert")
        :param primary_key: the primary key columns that define a unique record
        """
        if strategy != "insert" and (primary_key is None or len(primary_key) == 0):
            logger.error(
                "Cannot store using strategy {} when no primary_key is provided",
                strategy,
            )
            raise ValueError("Primary key must be provided when strategy is update or upsert")

        super().__init__(server, username, password, database)
        self._table = table
        self._strategy = strategy
        self._primary_key = primary_key

    def _insert(self, connection: pymssql.Connection, records: list[dict]) -> list[str]:
        """
        Insert records into the specified table.

        :param connection: the database connection to use
        :param records: the records to insert
        :return: a list of results, specifying `inserted` or an error or warning
        """
        logger.debug(
            "Inserting records into table {}: {}",
            self._table,
            records,
        )
        logger.info(
            "Inserting {} records into table {}",
            len(records),
            records,
        )

        results = []
        with connection.cursor() as cursor:
            for record in records:
                columns = record.keys()
                column_names = ", ".join(columns)

                value_placeholders = ", ".join(["%s"] * len(columns))

                q = f"INSERT INTO {self._table} ({column_names}) VALUES ({value_placeholders});"
                try:
                    logger.debug("Inserting record using query {}", q)
                    cursor.execute(q, [record[column] for column in columns])
                except pymssql.Error as e:
                    logger.warning("Failed to insert record, '{}'", str(e))
                    results.append(f"error {e.__class__.__name__}: {e.args}")
                except pymssql.Warning as e:
                    logger.warning("Warning with inserting record, '{}'", str(e))
                    results.append(f"warning {e.__class__.__name__}: {e.args}")
                else:
                    logger.debug("Inserted record")
                    results.append("inserted")

        return results

    def _check_missing_primary_key_columns(self, record: dict) -> list[str]:
        """
        Check what columns in the primary key are not included on the record.
        :param record: the record to check
        :return: a list of primary key columns that should be included in the record
        """
        return [
            col_name
            for col_name in self._primary_key
            if col_name not in record.keys()
        ]

    def _compile_assignments(self, record: dict[str, Any]) -> str:
        """
        Compile the assignments that should go on the SET part of the query

        :param record: the record to compile from
        :return: the string of assignments
        """
        assignments = [
            f"{col_name} = {quote_if_string(value)}"
            for col_name, value in record.items()
            if col_name not in self._primary_key
        ]
        return ", ".join(assignments)

    def _compile_where(self, record: dict[str, Any]) -> str:
        """
        Compile the WHERE clause to select the record to update.
        :param record: the record to compile from
        :return: the string of WHERE filter
        """
        where = [
            f"{col_name} = {quote_if_string(record[col_name])}"
            for col_name in self._primary_key
        ]
        return " AND ".join(where)

    def _update(self, connection: pymssql.Connection, records: list[dict]) -> list[str]:
        """
        Updating records in the configured table. Records should contain the values for
        the primary key columns. Values outside of these primary key columns are set to
        the values provided in the record.

        :param connection: the connection to the database
        :param records: the records to update
        :return: a list of results, specifying `updated` or an error or warning
        """
        logger.debug("Updating records in table {}: {}", self._table, records)
        logger.info("Updating {} records in table {}", len(records), self._table)

        result = []
        with connection.cursor() as cursor:
            for record in records:
                assignments = self._compile_assignments(record)
                if assignments == "":
                    logger.warning("No fields to update")
                    result.append("warning: no fields to update")
                    continue

                try:
                    where = self._compile_where(record)
                except KeyError:
                    missing_key_columns = ", ".join(
                        [
                            col_name
                            for col_name in self._primary_key
                            if col_name not in record.keys()
                        ]
                    )
                    logger.warning("Missing primary key columns '{}'", missing_key_columns)
                    result.append(f"warning: missing primary key columns: {missing_key_columns}")
                    continue

                q = f"UPDATE {self._table} SET {assignments} WHERE {where};"

                try:
                    logger.debug("Updating record using query {}", q)
                    cursor.execute(q)
                except pymssql.Error as e:
                    logger.warning("Failed to update record, '{}'", str(e))
                    result.append(f"error {e.__class__.__name__}: {e.args}")
                    continue
                except pymssql.Warning as e:
                    logger.warning("Warning with update record, '{}'", str(e))
                    result.append(f"warning {e.__class__.__name__}: {e.args}")

                if cursor.rowcount > 1:
                    logger.warning(
                        "Updated more than one records ({}) using primary key {}",
                        cursor.rowcount,
                        where,
                    )
                    result.append(f"warning: updated {cursor.rowcount} records")
                elif cursor.rowcount == 1:
                    logger.debug("Updated single record using primary key {}", where)
                    result.append("updated")
                else:
                    logger.warning("Updated no records using primary key {}", where)
                    result.append("missing")

        return result

    def _upsert(self, connection: pymssql.Connection, records: list[dict]) -> list[str]:
        """
        Upsert (Update or Insert) records into the configured table. Records should contain
        values for the primary key columns. Values outside of these primary key columns are
        set to the values provided in the record.

        :param connection: the connection to the database
        :param records: the records to upsert
        :return: a list of results, specifying `updated`, `inserted` or an error or warning
        """
        logger.debug("Upserting records in table {}: {}", self._table, records)
        logger.info("Upserting {} records in table {}", len(records), self._table)

        update_results = self._update(connection, records)
        records_to_insert = [
            IndexedRecord(index=index, record=record)
            for index, record in enumerate(records)
            if update_results[index] == "missing"
        ]
        if not records_to_insert:
            logger.debug("Upserted (only updates) records with results: {}", update_results)
            logger.info("No records to insert, only updated {} records", len(update_results))
            return update_results

        logger.info(
            "Upserting (inserts) {} records into table {}",
            len(records_to_insert),
            self._table,
        )
        insert_results = self._insert(
            connection,
            [indexed_record.record for indexed_record in records_to_insert]
        )
        logger.debug("Upsert inserting results: {}", insert_results)
        for insert_index, insert_result in enumerate(insert_results):
            original_index = records_to_insert[insert_index].index
            update_results[original_index] = insert_result

        return update_results

    def store(self, records: list[dict]) -> list[str]:
        """
        Store records into the configured table, using the configured strategy. Returns a
        list of results per record.

        :param records: the records to store
        :return: a list of results in the same order as the provided records.
        """
        if len(records) == 0:
            logger.warning("No records to store")
            return []

        with self._manager_factory.create_manager() as connection:
            match self._strategy:
                case "insert":
                    results = self._insert(connection, records)
                case "update":
                    results = self._update(connection, records)
                case "upsert":
                    results = self._upsert(connection, records)
                case _:
                    raise ValueError("Strategy must be insert, update or upsert")

            connection.commit()
            logger.debug(
                "Stored {} records into table {}, using strategy {}, with results: {}",
                len(records),
                self._table,
                self._strategy,
                results,
            )
            counter = collections.Counter(results)
            logger.info(
                "Stored {} records into table {}, using strategy {}, with result counts: {}",
                len(records),
                self._table,
                self._strategy,
                dict(counter),
            )
            return results



import pymssql

from conftest import MockMSSQLCursor, MockMSSQLConnection
from genie_flow_invoker.invoker.ms_sql_server.query import Querier


def test_syntax_error(monkeypatch):

    def connect(*args, **kwargs):
        cursor = MockMSSQLCursor(
            exception=pymssql.OperationalError("Syntax Error"),
        )
        return MockMSSQLConnection(cursor=cursor)

    monkeypatch.setattr(pymssql, "connect", connect)

    querier = Querier(
        "localhost",
        "some-user",
        "some-password",
        "some-database",
    )
    results = querier.query("SELECT AND INSERT SOME STUFF")

    assert results.results == []
    assert results.error == "error OperationalError: ('Syntax Error',)"


def test_no_results(monkeypatch):

    def connect(*args, **kwargs):
        cursor = MockMSSQLCursor(results={"SELECT * FROM tesy;": []})
        return MockMSSQLConnection(cursor=cursor)

    monkeypatch.setattr(pymssql, "connect", connect)

    querier = Querier(
        "localhost",
        "some-user",
        "some-password",
        "some-database",
    )
    results = querier.query("SELECT * FROM tesy;")

    assert results.results == []
    assert results.error is None


def test_query(monkeypatch):

    def connect(*args, **kwargs):
        cursor = MockMSSQLCursor(
            results={"SELECT * FROM tesy;": [dict(aap=10, noot=20)]}
        )
        return MockMSSQLConnection(cursor=cursor)

    monkeypatch.setattr(pymssql, "connect", connect)

    querier = Querier(
        "localhost",
        "some-user",
        "some-password",
        "some-database",
    )
    results = querier.query("SELECT * FROM tesy;")

    assert results.results == [dict(aap=10, noot=20)]
    assert results.error is None


def test_insert(monkeypatch):
    def connect(*args, **kwargs):
        cursor = MockMSSQLCursor(
            results={"INSERT INTO tesy (aap, noot) VALUES (10, 20);": []}
        )
        return MockMSSQLConnection(cursor=cursor)

    monkeypatch.setattr(pymssql, "connect", connect)

    querier = Querier(
        "localhost",
        "some-user",
        "some-password",
        "some-database",
    )
    results = querier.query("INSERT INTO tesy (aap, noot) VALUES (10, 20);")
    assert results.results == []



import json

import pymssql

from conftest import MockMSSQLCursor, MockMSSQLConnection
from genie_flow_invoker.invoker.ms_sql_server.store import Storer


def test_store_table_not_found(monkeypatch):

    def connect(*args, **kwargs):
        cursor = MockMSSQLCursor(
            exception=pymssql.OperationalError("Table not found"),
        )
        return MockMSSQLConnection(cursor)

    monkeypatch.setattr(pymssql, "connect", connect)

    storer = Storer(
        "localhost",
        "some-user",
        "some-pass",
        "some-database",
        "tesy",
        "insert",
    )
    result = storer.store([dict(aap=1)])

    assert result == ["error OperationalError: ('Table not found',)"]


def test_store_insert(monkeypatch):

    def connect(*args, **kwargs):
        cursor = MockMSSQLCursor(
            results={
                (
                    "INSERT INTO tesy (aap, noot) VALUES (%s, %s);",
                    json.dumps([(i, 10 * i)]),
                ): "inserted"
                for i in range(1, 3)
            }
        )
        return MockMSSQLConnection(cursor)

    monkeypatch.setattr(pymssql, "connect", connect)

    storer = Storer(
        "localhost",
        "some-user",
        "some-pass",
        "some-database",
        "tesy",
        "insert",
    )
    result = storer.store([dict(aap=1, noot=10), dict(aap=2, noot=20)])

    assert result == ["inserted", "inserted"]


def test_store_update(monkeypatch):

    def connect(*args, **kwargs):
        cursor = MockMSSQLCursor(
            results={
                "UPDATE tesy SET mies = 10 WHERE aap = 'noot';": ["updated"],
            },
            missing={
                "UPDATE tesy SET mies = 20 WHERE aap = 'wim';": ["missing"],
            },
        )
        return MockMSSQLConnection(cursor)

    monkeypatch.setattr(pymssql, "connect", connect)

    storer = Storer(
        "localhost",
        "some-user",
        "some-pass",
        "some-database",
        "tesy",
        "update",
        ["aap"],
    )
    result = storer.store([dict(aap="noot", mies=10), dict(aap="wim", mies=20)])

    assert result == ["updated", "missing"]


def test_store_update_missing_keys(monkeypatch):

    def connect(*args, **kwargs):
        cursor = MockMSSQLCursor(results={})
        return MockMSSQLConnection(cursor)

    monkeypatch.setattr(pymssql, "connect", connect)

    storer = Storer(
        "localhost",
        "some-user",
        "some-pass",
        "some-database",
        "tesy",
        "update",
        primary_key=["aap", "noot"],
    )
    result = storer.store([dict(wim=10)])

    assert result == ["warning: missing primary key columns: aap, noot"]


def test_store_update_multi(monkeypatch):

    def connect(*args, **kwargs):
        cursor = MockMSSQLCursor(
            results={
                "UPDATE tesy SET mies = 10 WHERE aap = 'noot';": ["updated", "updated"],
            }
        )
        return MockMSSQLConnection(cursor)

    monkeypatch.setattr(pymssql, "connect", connect)

    storer = Storer(
        "localhost",
        "some-user",
        "some-pass",
        "some-database",
        "tesy",
        "update",
        primary_key=["aap"],
    )
    result = storer.store([dict(aap="noot", mies=10)])

    assert result == ["warning: updated 2 records"]


def test_store_upsert(monkeypatch):

    def connect(*args, **kwargs):
        cursor = MockMSSQLCursor(
            results={
                "UPDATE tesy SET mies = 10 WHERE aap = 'noot';": ["updated"],
                (
                    "INSERT INTO tesy (aap, mies) VALUES (%s, %s);",
                    json.dumps([("wim", 20)]),
                ): ["inserted"],
            }
        )
        return MockMSSQLConnection(cursor)

    monkeypatch.setattr(pymssql, "connect", connect)

    storer = Storer(
        "localhost",
        "some-user",
        "some-pass",
        "some-database",
        "tesy",
        "upsert",
        primary_key=["aap"],
    )
    result = storer.store([dict(aap="noot", mies=10), dict(aap="wim", mies=20)])

    assert result == ["updated", "inserted"]



import pymssql

from conftest import MockMSSQLCursor, MockMSSQLConnection
from genie_flow_invoker.invoker.ms_sql_server.retrieve import Retriever


def test_retrieve_table_not_found(monkeypatch):

    def connect(*args, **kwargs):
        cursor = MockMSSQLCursor(
            exception=pymssql.OperationalError("table not found"),
        )
        return MockMSSQLConnection(cursor)

    monkeypatch.setattr(pymssql, "connect", connect)
    retriever = Retriever(
        "localhost",
        "some-user",
        "some-pass",
        "some-database",
        "tesy",
    )
    results = retriever.retrieve()

    assert results.error == "error OperationalError: ('table not found',)"
    assert results.results == []


def test_retrieve_all(monkeypatch):
    expected_results = [dict(aap=1, noot=2), dict(aap=3, noot=4)]

    def connect(*arg, **kwargs):
        cursor = MockMSSQLCursor(
            results={
                "SELECT * FROM tesy;": expected_results,
            }
        )
        return MockMSSQLConnection(cursor)

    monkeypatch.setattr(pymssql, "connect", connect)

    retriever = Retriever(
        "localhost",
        "some-user",
        "some-pass",
        "some-database",
        "tesy",
    )
    results = retriever.retrieve()

    assert results.error is None
    assert results.results == expected_results


def test_retrieve_top(monkeypatch):
    expected_results = [dict(aap=1, noot=2), dict(aap=3, noot=4)]

    def connect(*args, **kwargs):
        cursor = MockMSSQLCursor(
            results={
                "SELECT TOP 2 * FROM tesy;": expected_results,
            }
        )
        return MockMSSQLConnection(cursor)

    monkeypatch.setattr(pymssql, "connect", connect)

    retriever = Retriever(
        "localhost",
        "some-user",
        "some-pass",
        "some-database",
        "tesy",
        top=2,
    )
    results = retriever.retrieve()

    assert results.error is None
    assert results.results == expected_results


def test_retrieve_order(monkeypatch):
    expected_results = [dict(aap=1, noot=2), dict(aap=3, noot=4)]

    def connect(*args, **kwargs):
        cursor = MockMSSQLCursor(
            results={
                "SELECT * FROM tesy ORDER BY aap, noot DESC;": expected_results,
            }
        )
        return MockMSSQLConnection(cursor)

    monkeypatch.setattr(pymssql, "connect", connect)

    retriever = Retriever(
        "localhost",
        "some-user",
        "some-pass",
        "some-database",
        "tesy",
        sort_order=["aap", "-noot"],
    )
    results = retriever.retrieve()

    assert results.error is None
    assert results.results == expected_results


def test_retrieve_top_and_order(monkeypatch):
    expected_results = [dict(aap=1, noot=2), dict(aap=3, noot=4)]

    def connect(*args, **kwargs):
        cursor = MockMSSQLCursor(
            results={
                "SELECT TOP 2 * FROM tesy ORDER BY aap;": expected_results,
            }
        )
        return MockMSSQLConnection(cursor)

    monkeypatch.setattr(pymssql, "connect", connect)

    retriever = Retriever(
        "localhost",
        "some-user",
        "some-pass",
        "some-database",
        "tesy",
        top=2,
        sort_order=["aap"],
    )
    results = retriever.retrieve()

    assert results.error is None
    assert results.results == expected_results



from genie_flow_invoker.invoker.ms_sql_server import (
    MSSQLServerRetrieveInvoker,
    MSSQLServerStoreInvoker,
    MSSQLServerQueryInvoker,
)


def test_retrieve_invoker_env_vars(monkeypatch):
    monkeypatch.setenv("MS_SQL_SERVER", "patched-localhost")
    monkeypatch.setenv("MS_SQL_SERVER_USERNAME", "patched-username")
    monkeypatch.setenv("MS_SQL_SERVER_PASSWORD", "patched-password")
    monkeypatch.setenv("MS_SQL_SERVER_DATABASE", "patched-database")
    monkeypatch.setenv("MS_SQL_SERVER_TABLE", "patched-table")

    retriever = MSSQLServerRetrieveInvoker.from_config(dict())

    assert retriever.base_config["server"] == "patched-localhost"
    assert retriever.base_config["username"] == "patched-username"
    assert retriever.base_config["password"] == "patched-password"
    assert retriever.base_config["database"] == "patched-database"
    assert retriever.base_config["table"] == "patched-table"
    assert retriever.base_config["top"] is None
    assert retriever.base_config["sort_order"] is None


def test_retrieve_invoker_mix(monkeypatch):
    monkeypatch.setenv("MS_SQL_SERVER", "patched-localhost")
    monkeypatch.setenv("MS_SQL_SERVER_USERNAME", "patched-username")
    monkeypatch.setenv("MS_SQL_SERVER_PASSWORD", "patched-password")
    monkeypatch.setenv("MS_SQL_SERVER_DATABASE", "patched-database")

    retriever = MSSQLServerRetrieveInvoker.from_config(
        dict(table="set-table", top=25, sort_order=["-aap", "noot"])
    )

    assert retriever.base_config["server"] == "patched-localhost"
    assert retriever.base_config["username"] == "patched-username"
    assert retriever.base_config["password"] == "patched-password"
    assert retriever.base_config["database"] == "patched-database"
    assert retriever.base_config["table"] == "set-table"
    assert retriever.base_config["top"] == 25
    assert retriever.base_config["sort_order"] == ["-aap", "noot"]


class MSQSQLServerStoreInvoker:
    pass


def test_store_invoker_env_vars(monkeypatch):
    monkeypatch.setenv("MS_SQL_SERVER", "patched-localhost")
    monkeypatch.setenv("MS_SQL_SERVER_USERNAME", "patched-username")
    monkeypatch.setenv("MS_SQL_SERVER_PASSWORD", "patched-password")
    monkeypatch.setenv("MS_SQL_SERVER_DATABASE", "patched-database")
    monkeypatch.setenv("MS_SQL_SERVER_TABLE", "patched-table")

    storer = MSSQLServerStoreInvoker.from_config(dict())

    assert storer.base_config["server"] == "patched-localhost"
    assert storer.base_config["username"] == "patched-username"
    assert storer.base_config["password"] == "patched-password"
    assert storer.base_config["database"] == "patched-database"
    assert storer.base_config["table"] == "patched-table"


def test_store_invoker_mix(monkeypatch):
    monkeypatch.setenv("MS_SQL_SERVER", "patched-localhost")
    monkeypatch.setenv("MS_SQL_SERVER_USERNAME", "patched-username")
    monkeypatch.setenv("MS_SQL_SERVER_PASSWORD", "patched-password")
    monkeypatch.setenv("MS_SQL_SERVER_DATABASE", "patched-database")

    storer = MSSQLServerStoreInvoker.from_config(
        dict(
            table="set-table",
            strategy="upsert",
            primary_key=["aap", "noot"],
        )
    )

    assert storer.base_config["server"] == "patched-localhost"
    assert storer.base_config["username"] == "patched-username"
    assert storer.base_config["password"] == "patched-password"
    assert storer.base_config["database"] == "patched-database"
    assert storer.base_config["table"] == "set-table"
    assert storer.base_config["strategy"] == "upsert"
    assert storer.base_config["primary_key"] == ["aap", "noot"]


def test_query_invoker_env_vars(monkeypatch):
    monkeypatch.setenv("MS_SQL_SERVER", "patched-localhost")
    monkeypatch.setenv("MS_SQL_SERVER_USERNAME", "patched-username")
    monkeypatch.setenv("MS_SQL_SERVER_PASSWORD", "patched-password")
    monkeypatch.setenv("MS_SQL_SERVER_DATABASE", "patched-database")

    querier = MSSQLServerQueryInvoker.from_config(dict())

    assert querier.base_config["server"] == "patched-localhost"
    assert querier.base_config["username"] == "patched-username"
    assert querier.base_config["password"] == "patched-password"
    assert querier.base_config["database"] == "patched-database"



from typing import Optional

from pydantic import BaseModel, Field


class RetrievedResults(BaseModel):
    """
    The model for returned records and / errors.
    """
    results: list[dict] = Field(
        default_factory=list,
        description="A list of dictionaries returned by a query",
    )
    error: Optional[str] = Field(
        default=None,
        description="The optional error message returned by a query",
    )



from typing import Optional

from loguru import logger
import pymssql

from .manager import MSSQLServerOperator
from .model import RetrievedResults


class Retriever(MSSQLServerOperator):

    def __init__(
            self,
            server: str,
            username: str,
            password: str,
            database: str,
            table: str,
            top: Optional[int] = None,
            sort_order: Optional[list[str]] = None,
    ):
        super().__init__(server, username, password, database)
        self._table = table
        self._top = top
        self._sort_order = sort_order

    def _compile_query(self) -> str:
        top = f" TOP {self._top}" if self._top is not None else ""
        logger.debug(
            "Compiling query for top set at {}, with top statement '{}'",
            self._top,
            top,
        )

        sort_order = ""
        if self._sort_order is not None and len(self._sort_order) > 0:
            order = [
                f"{sort_col[1:]} DESC" if sort_col.startswith("-") else sort_col
                for sort_col in self._sort_order
            ]
            sort_order = " ORDER BY " + ", ".join(order)
            logger.debug(
                "Compiling query for sort order {} with order by statement '{}'",
                self._sort_order,
                sort_order,
            )

        return f"SELECT{top} * FROM {self._table}{sort_order};"

    def retrieve(self) -> RetrievedResults:
        """
        Retrieve the records as configured.

        :return: the retrieved records object that contains the records and / or error
        messages
        """
        logger.info(
            "Retrieving from table {} the top {} rows, ordered by {}",
            self._table,
            self._top,
            self._sort_order,
        )
        with self._manager_factory.create_manager() as connection:
            q = self._compile_query()
            logger.debug("Retrieval query: {}", q)
            with connection.cursor(as_dict=True) as cursor:
                try:
                    cursor.execute(q)
                except pymssql.Error as e:
                    logger.warning("Retrieve query failed, {}", str(e))
                    return RetrievedResults(
                        error=f"error {e.__class__.__name__}: {e.args}",
                    )
                except pymssql.Warning as e:
                    logger.warning(
                        "Retrieve query warning {} with {} results",
                        str(e),
                        cursor.rowcount,
                    )
                    return RetrievedResults(
                        results=cursor.fetchall() if cursor.rowcount > 0 else [],
                        error=f"warning {e.__class__.__name__}: {e.args}",
                    )
                else:
                    return RetrievedResults(
                        results=cursor.fetchall(),
                    )
